import json
import pickle
from pathlib import Path
from typing import Any, Dict, List

import av
import av.container
import av.video
import numpy as np
import yaml

from ..math.geometry import rotmat_to_quaternion
from ..system import logger

__all__ = [
    'write_json', 'write_text', 'write_bytes', 'write_yaml', 'write_pickle',
    'write_image', 'write_points_pcd', 'VideoWriter'
]


def _check_suffix(filepath: str, suffix: str) -> Path:
    """
    Ensure output filename with correct suffix

    return {filepath}.{suffix}
    """
    filepath_pth = Path(filepath)
    if not filepath_pth.suffix or filepath_pth.suffix != suffix:
        filepath = '{}{}'.format(filepath, suffix)
    return Path(filepath)


def write_text(filepath: str, output: Any, verbose: bool = False) -> Any:
    if verbose:
        logger.info(f'write texts file {filepath}')
    return Path(filepath).write_text(output)


def write_bytes(filepath: str, output: Any, verbose: bool = False) -> Any:
    if verbose:
        logger.info(f'write bytes file {filepath}')
    return Path(filepath).write_bytes(output)


def write_json(filepath: str, jsondict: Dict, verbose: bool = False) -> Any:
    _filepath = _check_suffix(filepath, '.json')
    if verbose:
        logger.info(f'write json file {_filepath}')
    return _filepath.write_text(json.dumps(jsondict, indent=4, sort_keys=False))


def write_yaml(filepath: str, yamldict: Dict, verbose: bool = False) -> Any:
    """
    Read yaml

    @param filepath
    """
    _filepath = _check_suffix(filepath, '.yaml')
    if verbose:
        logger.info(f'write yaml file {_filepath}')
    with open(_filepath, 'w') as f:
        yaml.dump(yamldict, f)


def write_pickle(filepath: str, output: Any, verbose: bool = False) -> None:
    _filepath = _check_suffix(filepath, '.pkl')
    if verbose:
        logger.info(f'write pickle file {_filepath}')
    with open(_filepath, 'wb') as f:
        return pickle.dump(output, f)


def write_image(filepath: str, output: Any):
    import cv2
    return cv2.imwrite(filepath, output)


class VideoWriter:

    def __init__(self, filepath: str, fps: int = 10):
        self.filepath = _check_suffix(filepath, '.mp4')

        self.fps = fps
        self.is_closed = False

    def __create_container(self, img_w, img_h):
        self.filepath.parent.mkdir(parents=True, exist_ok=True)
        self.container = av.container.open(self.filepath, mode='w')
        self.stream = self.container.add_stream(
            codec_name='libx264', rate=self.fps,
            options=dict(crf='23', maxrate='16M', bufsize='32M')
        )
        self.stream.pix_fmt = 'yuv420p'
        self.stream.width = img_w
        self.stream.height = img_h

    def write(self, image: np.ndarray):
        if not hasattr(self, 'container'):
            H, W, _ = image.shape
            self.__create_container(img_w=W, img_h=H)

        frame = av.video.frame.VideoFrame.from_ndarray(image, format="bgr24")
        for packet in self.stream.encode(frame):
            self.container.mux(packet)

    def close(self):
        # do not call close duplicated
        if not hasattr(self, 'container') or self.is_closed:
            return

        for packet in self.stream.encode():
            self.container.mux(packet)

        self.container.close()
        self.is_closed = True

    def __del__(self):
        self.close()


def write_points_pcd(
    filepath: str,
    output: np.ndarray,
    fields: List = ['x', 'y', 'z', 'intensity', 'timestamp'],
    viewpoint: List = None,
    verbose: bool = False
):
    _filepath = _check_suffix(filepath, '.pcd')
    from pypcd.pypcd import (PointCloud, numpy_type_to_pcd_type,
                             save_point_cloud_bin)

    field_map = {
        'x': np.dtype('float32'),
        'y': np.dtype('float32'),
        'z': np.dtype('float32'),
        'intensity': np.dtype('uint8'),
        'timestamp': np.dtype('float64'),
        'ring': np.dtype('uint8'),
    }
    pts_metadata = {
        'version': .7,
        'fields': fields,
        'size': [numpy_type_to_pcd_type[field_map[fname]][1] for fname in fields],
        'type': [numpy_type_to_pcd_type[field_map[fname]][0] for fname in fields],
        'count': [1 for _ in fields],
        'width': len(output),
        'height': 1,
        'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
        'points': len(output),
        'data': 'binary',
    }

    if viewpoint is not None:
        pts_metadata.update(viewpoint=np.hstack([
            viewpoint[:3, 3], rotmat_to_quaternion(viewpoint[:3, :3])]).tolist())

    pc_data = np.asarray([tuple(p) for p in output],
        dtype=np.dtype([(f, field_map[f]) for f in fields]))

    pts_obj = PointCloud(pts_metadata, pc_data)
    save_point_cloud_bin(pts_obj, _filepath)
    if verbose:
        logger.info(f'write points file {_filepath}')



from functools import wraps
from pathlib import Path
from typing import Any, Callable, List

import numpy as np

from ..math.geometry import quaternion_to_rotmat, transform_matrix
from ..system.logging import logger

__all__ = [
    'read_pickle', 'read_json', 'read_yaml', 'read_text', 'read_bytes',
    'read_image', 'read_numpy_bin', 'read_points_pcd', 'check_filepath'
]


def check_filepath(func: Callable, *args, **kwargs):
    """
    Ensure filepath type & check if the file exist

    @func (Callable[str]): The function to decorate
    return Callable[Path]: A wrapper function.
    """
    def _check_filepath(filepath: str, *, strict: bool = True, no_warning: bool = False) -> Path:
        _filepath = filepath if isinstance(filepath, Path) else Path(filepath)
        if _filepath.exists():
            return _filepath
        elif not no_warning:
            logger.warning(f"The path {_filepath} does not exist.")
        if strict:
            raise FileNotFoundError(f"The path {filepath} does not exist.")

    @wraps(func)
    def wrapper(filepath: str, *args, strict: bool = True, verbose: bool = False, no_warning: bool = False, **kwargs):
        _filepath = _check_filepath(filepath, strict=strict, no_warning=no_warning)
        if verbose:
            logger.debug('Read file', _filepath.absolute())
        if _filepath is None:
            return None
        return func(_filepath, *args, **kwargs)

    if isinstance(func, (str, Path)):
        return _check_filepath(func, *args, **kwargs)
    return wrapper


@check_filepath
def read_pickle(filepath: Path) -> Any:
    """
    Read content from a pickle file

    @param filepath: path to pickle file
    """
    from pickle import load
    with open(filepath, 'rb') as f:
        pkl_dict = load(f)
    return pkl_dict


@check_filepath
def read_json(filepath: Path) -> Any:
    """
    Read jsondict

    @param filepath: path to json file
    """
    from json import loads
    return loads(filepath.read_text())


@check_filepath
def read_text(filepath: Path) -> Any:
    """
    Read plain ascii texts

    @param filepath
    """
    return filepath.read_text().splitlines()


@check_filepath
def read_bytes(filepath: Path) -> Any:
    """
    Read file in bytes

    @param filepath
    """
    return filepath.read_bytes().splitlines()


@check_filepath
def read_yaml(filepath: Path) -> Any:
    """
    Read yaml
    
    @param filepath
    """
    import yaml
    return yaml.safe_load(filepath.read_text())


@check_filepath
def read_numpy_bin(filepath: Path, dtype: type = np.float32) -> np.ndarray:
    return np.fromfile(filepath, dtype=dtype)


@check_filepath
def read_image(filepath: Path, flags: int = None) -> np.ndarray:
    import cv2
    return cv2.imread(str(filepath), flags=flags)


@check_filepath
def read_points_pcd(
    filepath: Path,
    fields: List = ['x', 'y', 'z', 'intensity', 'timestamp'],
    viewpoint: bool = False,
) -> np.ndarray:
    from pypcd import pypcd
    pcd = pypcd.point_cloud_from_path(filepath)

    pcd_metadata = pcd.get_metadata()
    data_length = len([_ for _ in pcd_metadata['fields'] if _ != '_' and _ in fields])
    data_np = np.zeros((pcd_metadata['width'], data_length), dtype=float)

    for name in pcd_metadata['fields']:
        if name not in fields:
            continue
        data_np[:, fields.index(name)] = pcd.pc_data[name]

    if viewpoint:
        return data_np, transform_matrix(
            pcd_metadata['viewpoint'][:3],
            quaternion_to_rotmat(pcd_metadata['viewpoint'][3:]))

    return data_np



from pathlib import Path
from shutil import copy, copytree, rmtree
from typing import List

import numpy as np

from ..system.logging import logger
from .read_file import *
from .write_file import *


def read_points(filepath: Path, **kwargs) -> np.ndarray:
    """ """
    _filepath = filepath if isinstance(filepath, Path) else Path(filepath)
    ext_dict = {
        '.pcd': read_points_pcd,
    }
    return ext_dict[Path(_filepath).suffix](_filepath, **kwargs)


def mkdir(dirpath: Path, verbose: bool = False) -> Path:
    _dirpath = dirpath if isinstance(dirpath, Path) else Path(dirpath)
    if not _dirpath.exists():
        _dirpath.mkdir(parents=True, exist_ok=True)
        if verbose:
            logger.info(f'make directory {_dirpath.resolve()}')
    return _dirpath


def lsdir(dirpath: Path, pattern: str = '*', return_str=False) -> List[Path]:
    _dirpath = dirpath if isinstance(dirpath, Path) else Path(dirpath)
    path_type = str if return_str else Path
    return sorted([path_type(n) for n in _dirpath.glob(pattern)])


def emptydir(dirpath: Path, verbose: bool = True) -> Path:
    _dirpath = dirpath if isinstance(dirpath, Path) else Path(dirpath)
    if verbose:
        logger.info(f'empty directory {_dirpath}')
    if not _dirpath.exists():
        _dirpath = mkdir(_dirpath, exist_ok=True)

    for _dir in lsdir(_dirpath):
        if _dir.is_symlink() or _dir.is_file():
            _dir.unlink()
        elif _dir.is_dir():
            rmdir(_dir, verbose=False)
    return _dirpath


def rmdir(dirpath: Path, verbose: bool = False) -> Path:
    _dirpath = dirpath if isinstance(dirpath, Path) else Path(dirpath)
    if _dirpath.exists():
        if verbose:
            logger.info(f'remove directory {_dirpath}')
        if _dirpath.is_dir():
            emptydir(_dirpath, verbose=False)
            rmtree(_dirpath)
        else:
            _dirpath.unlink()
    return _dirpath


def cpdir(dirpath: Path, tgtpath: Path, verbose: bool = False) -> Path:
    _dirpath = dirpath if isinstance(dirpath, Path) else Path(dirpath)
    _tgtpath = tgtpath if isinstance(tgtpath, Path) else Path(tgtpath)
    if _dirpath.exists():
        if verbose:
            logger.info(f'copy filepath {_dirpath} to {tgtpath}')
        if _tgtpath.exists():
            rmdir(_tgtpath)
        if _dirpath.is_dir():
            copytree(_dirpath, tgtpath)
        else:
            copy(_dirpath, tgtpath)
    return _tgtpath


from setuptools import setup, find_packages

setup(
    name='omnyx',
    version='0.0.1',
    description=(),
    long_description=(),
    author='tongbochen',
    author_email='bochen.tong@proton.me',
    url='',
    license='MIT Licence',
    packages=find_packages(),
    install_requires=[],
)
